
#include "asm_passes.h"
#include "../arm.h"

// Optimization passes
#include "stack_ra.h"
#include "register_allocation.h"
#include "simplify_asm.h"
#include "remove_identical_moves.h"
#include "block_placement.h"
#include "remove_redundant_ldrs.h"
#include "condify.h"

inline void run_on_every_function(Program_Asm *program_asm, Asm_Func_Pass pass) {
    for(auto func_asm : program_asm->functions) {
        pass(func_asm);
    }
}

#define RUN(pass_name) pass_name(program_asm)
void bless(Program_Asm *program_asm, bool opt) {
	RUN(remove_redundant_ldrs);
    register_allocation(program_asm, opt);
    sregister_allocation(program_asm, opt);
    RUN(remove_identical_moves);
    printf("ok\n");
    // RUN(stack_ra);
    if (opt) {
        RUN(simplify_asm);
        RUN(remove_redundant_ldrs);
        RUN(block_placement);
        RUN(condify);
    }
}
#undef RUN

// helper functions

void replace_uses(MI *I, MOperand old_opr, MOperand new_opr) {
    Array<MOperand *> oprs;

    switch(I->tag) {

        case MI_MOVE:   {
            oprs.push(&((MI_Move *)I)->src);
        } break;

        case MI_CLZ:   {
            oprs.push(&((MI_Clz *)I)->operand);
        } break;

        case MI_BINARY: {
            oprs.push(&((MI_Binary *)I)->lhs);
            oprs.push(&((MI_Binary *)I)->rhs);
        } break;

        case MI_COMPARE: {
            oprs.push(&((MI_Compare *)I)->lhs);
            oprs.push(&((MI_Compare *)I)->rhs);
        } break;

        case MI_BRANCH: {
            oprs.push(&((MI_Branch *)I)->lhs);
            oprs.push(&((MI_Branch *)I)->rhs);
        } break;

        // no need to handle function call
        case MI_FUNC_CALL: {} break;

        case MI_STORE: { 
            oprs.push(&((MI_Store *)I)->reg);
            oprs.push(&((MI_Store *)I)->base);
            oprs.push(&((MI_Store *)I)->offset);
        } break;

        case MI_LOAD: { 
            oprs.push(&((MI_Load *)I)->base);
            oprs.push(&((MI_Load *)I)->offset);
        } break;

        case MI_VSTORE: {
            oprs.push(&((MI_VStore *)I)->base);
            oprs.push(&((MI_VStore *)I)->offset);
        } break;
        
        case MI_VLOAD: {
            oprs.push(&((MI_VLoad *)I)->base);
            oprs.push(&((MI_VLoad *)I)->offset);
        } break;

        case MI_VCVT: {
            if (((MI_VCvt *)I)->from_type == S32) {
                oprs.push(&((MI_VCvt *)I)->src);
            }
        } break;
        /*
        case MI_PUSH: {
            auto push = (MI_Push *) I;
            for(int i = 0; i < push->operands.len; i++) {
                oprs.push(&push->operands[i]);
            }
        } break;
        */

        // also no need for return
        case MI_RETURN:
        case MI_POP:
        case MI_PUSH:
        //----float----
        case MI_VBINARY:
        case MI_VCMP_SET:
        case MI_VPUSH:
        case MI_VPOP:
        break;
        default: {
            printf("unknown MI tag in replace_uses: %d\n", I->tag);
            assert(false);
        }

    }

    for(MOperand *opr_ptr : oprs) {
        if (opr_ptr && *opr_ptr == old_opr) {
            // @NOTE! remember to keep lsl info!!!
            opr_ptr->tag   = new_opr.tag;
            opr_ptr->value = new_opr.value;
            opr_ptr->adr   = new_opr.adr;
            //*opr_ptr = new_opr;
        }
    }

}

void s_replace_uses(MI *I, MOperand old_opr, MOperand new_opr) {
    Array<MOperand *> oprs;

    switch(I->tag) {

        case MI_VMOVE:   {
            oprs.push(&((MI_Move *)I)->src);
        } break;

        case MI_CLZ:   {
            oprs.push(&((MI_Clz *)I)->operand);
        } break;

        case MI_VBINARY: {
            oprs.push(&((MI_Binary *)I)->lhs);
            oprs.push(&((MI_Binary *)I)->rhs);
        } break;

        case MI_VCMP_SET: {
            oprs.push(&((MI_Compare *)I)->lhs);
            oprs.push(&((MI_Compare *)I)->rhs);
        } break;

        // no need to handle function call
        case MI_FUNC_CALL: {} break;

        case MI_VSTORE: { 
            oprs.push(&((MI_Store *)I)->reg);
        } break;

        case MI_VLOAD: {} break;

        case MI_VCVT: {
            if (((MI_VCvt *)I)->from_type == F32) {
                oprs.push(&((MI_VCvt *)I)->src);
            }
        } break;

        
        /*
        case MI_PUSH: {
            auto push = (MI_Push *) I;
            for(int i = 0; i < push->operands.len; i++) {
                oprs.push(&push->operands[i]);
            }
        } break;
        */

        case MI_MOVE:
        case MI_BINARY:
        case MI_COMPARE:
        case MI_LOAD:
        case MI_STORE:
        // also no need for return
        case MI_RETURN:
        case MI_POP:
        case MI_PUSH:
        case MI_BRANCH: 
        //----float----
        case MI_VPUSH:
        case MI_VPOP:
        break;
        default: {
            printf("unknown MI tag in s_replace_uses: %d\n", I->tag);
            assert(false);
        }

    }

    for(MOperand *opr_ptr : oprs) {
        if (opr_ptr && *opr_ptr == old_opr) {
            // @NOTE! remember to keep lsl info!!!
            opr_ptr->tag   = new_opr.tag;
            opr_ptr->value = new_opr.value;
            opr_ptr->adr   = new_opr.adr;
            //*opr_ptr = new_opr;
        }
    }

}

void replace_defs(MI *I, MOperand old_opr, MOperand new_opr) {
    Array<MOperand *> oprs;
    switch(I->tag) {
        case MI_MOVE:   { oprs.push(&((MI_Move *)I)->dst); } break;
        case MI_BINARY: { oprs.push(&((MI_Binary *)I)->dst); } break;

        // no need to do function call
        case MI_FUNC_CALL: {
        } break;

        // bx lr
        case MI_RETURN: {} break;

        case MI_LOAD: { oprs.push(&((MI_Load *)I)->reg); } break;

        case MI_CLZ: { oprs.push(&((MI_Clz *)I)->dst); } break;

        case MI_VCMP_SET: { oprs.push(&((MI_VCmp_Set *)I)->dst); } break;

        case MI_VCVT: {
            if (((MI_VCvt *)I)->to_type == S32) {
                oprs.push(&((MI_VCvt *)I)->dst);
            }
        } break;

        /*
        case MI_POP:  {
            auto pop = (MI_Pop *) I;
            for(int i = 0; i < pop->operands.len; i++) {
                oprs.push(&pop->operands[i]);
            }
        } break;
        */

        case MI_COMPARE:
        case MI_PUSH:
        case MI_POP:
        case MI_STORE: // @TODO: store/load may use self increment thing?
        case MI_BRANCH: 
        //----float----
        case MI_VBINARY:
        
        case MI_VPUSH:
        case MI_VPOP:
        case MI_VLOAD: // vldr
        case MI_VSTORE:  // vstr
        break;
        
        default: {
            printf("unknown MI tag in replace_defs: %d\n", I->tag);
            assert(false);
        } 
    }

    for(MOperand *opr_ptr : oprs) {
        if (opr_ptr && *opr_ptr == old_opr) {
            *opr_ptr = new_opr;
        }
    }

}

void s_replace_defs(MI *I, MOperand old_opr, MOperand new_opr) {
    Array<MOperand *> oprs;
    switch(I->tag) {
        case MI_VMOVE:   { oprs.push(&((MI_Move *)I)->dst); } break;
        case MI_VBINARY: { oprs.push(&((MI_Binary *)I)->dst); } break;

        // no need to do function call
        case MI_FUNC_CALL: {} break;

        case MI_RETURN: {} break;

        case MI_VLOAD: { oprs.push(&((MI_Load *)I)->reg); } break;

        case MI_CLZ: { oprs.push(&((MI_Clz *)I)->dst); } break;

        case MI_VCVT: {
            if (((MI_VCvt *)I)->to_type == F32) {
                oprs.push(&((MI_VCvt *)I)->dst);
            }
        } break;

        case MI_MOVE:
        case MI_BINARY:
        case MI_COMPARE:
        case MI_PUSH:
        case MI_POP:
        case MI_STORE: // @TODO: store/load may use self increment thing?
        case MI_BRANCH: 
        //----float----
        case MI_VCMP_SET:
        case MI_VPUSH:
        case MI_VPOP:
        case MI_VSTORE:  // vstr
        break;
        
        default: {
            printf("unknown MI tag in s_replace_defs: %d\n", I->tag);
            assert(false);
        } 
    }

    for(MOperand *opr_ptr : oprs) {
        if (opr_ptr && *opr_ptr == old_opr) {
            *opr_ptr = new_opr;
        }
    }

}

Array<MOperand> get_defs(MI *I) {
    Array<MOperand> oprs;
    switch(I->tag) {
        case MI_MOVE:   { oprs.push(((MI_Move *)I)->dst); } break;
        case MI_BINARY: { oprs.push(((MI_Binary *)I)->dst); } break;
        case MI_FUNC_CALL: {
            oprs.push(make_reg(ra));
            for(uint8 r = x1; r < REG_COUNT; r++) {
                if (is_caller_save(r)) {
                    if (r == ra) cout << "ra appear in function call" << endl;
                    oprs.push(make_reg(r));
                }
            }
        } break;

        case MI_LOAD: { oprs.push(((MI_Load *)I)->reg); } break;

        case MI_CLZ: { oprs.push(((MI_Clz *)I)->dst); } break;

        case MI_SLT: { oprs.push(((MI_Slt *)I)->dst); } break;

        case MI_VCVT: {
            auto fcvt = (MI_VCvt *)I;
            if(fcvt->to_type == S32) {
                oprs.push(fcvt->dst);
            }
        } break;

        case MI_VCMP_SET: {
            oprs.push(((MI_VCmp_Set *)I)->dst);
        } break;

        /*
        case MI_POP:  {
            auto pop = (MI_Pop *) I;
            for (auto pop_operand : pop->operands) {
                oprs.push(pop_operand);
            }
        } break;
        */

        case MI_COMPARE:
        case MI_PUSH:
        case MI_RETURN:
        case MI_POP:
        case MI_STORE: // @TODO: store/load may use self increment thing?
        case MI_BRANCH: 
        //----float----
        case MI_VMOVE:
        case MI_VBINARY:
        case MI_VPUSH:
        case MI_VPOP:
        case MI_VLOAD: // vldr
        case MI_VSTORE:  // vstr
        break;
        
        default: {
            printf("unknown instrID: %d\n", I->tag);
            assert(false && "meet unknown instr in get_defs");
        }
    }
    return oprs;
}

Array<MOperand> get_uses(MI *I, bool func_has_return_value) {
    Array<MOperand> oprs;

    switch(I->tag) {

        case MI_MOVE:   {
            oprs.push(((MI_Move *)I)->src);
        } break;

        case MI_VMOVE: {
            if(((MI_VMove *)I)->src.tag == REG || ((MI_VMove *)I)->src.tag == VREG){
                oprs.push(((MI_VMove *)I)->src);
            }
        } break;
        case MI_CLZ:   {
            oprs.push(((MI_Clz *)I)->operand);
        } break;

        case MI_BINARY: {
            oprs.push(((MI_Binary *)I)->lhs);
            oprs.push(((MI_Binary *)I)->rhs);
        } break;

        case MI_COMPARE: {
            oprs.push(((MI_Compare *)I)->lhs);
            oprs.push(((MI_Compare *)I)->rhs);
        } break;

        case MI_SLT: {
            oprs.push(((MI_Slt *)I)->lhs);
            oprs.push(((MI_Slt *)I)->rhs);
        } break;

        case MI_FUNC_CALL: {
            auto call = (MI_Func_Call *)I;
            // @TODO: get uses based on arg count
            // @TODO: temporarily use 4 argument regs to keep compatibility
            for(uint8 r = a0; r <= a3; r++) {
                if (r >= call->arg_count + a0) break;
                oprs.push(make_reg(r));
            }
        } break;

        case MI_STORE: { 
            oprs.push(((MI_Store *)I)->reg);
            oprs.push(((MI_Store *)I)->base);
            oprs.push(((MI_Store *)I)->offset);
        } break;

        case MI_LOAD: { 
            oprs.push(((MI_Load *)I)->base);
            oprs.push(((MI_Load *)I)->offset);
        } break;

        /*
        case MI_PUSH: {
            auto push = (MI_Push *) I;
            for (auto push_operand : push->operands) {
                oprs.push(push_operand);
            }
        } break;
        */

        case MI_RETURN: {
            oprs.push(make_reg(ra));
            if (func_has_return_value) {
                oprs.push(make_reg(a0));
            }
        } break;

        case MI_VLOAD:{
            auto vldr = (MI_VLoad *)I;
            if(vldr->base.tag == REG || vldr->base.tag == VREG){
                oprs.push(vldr->base);
            }
            if(vldr->offset.tag == REG || vldr->offset.tag == VREG){
                oprs.push(vldr->offset);
            }
        } break;
        case MI_VSTORE:{
            auto vstr = (MI_VStore *)I;
            if(vstr->base.tag == REG || vstr->base.tag == VREG){
                oprs.push(vstr->base);
            }
            if(vstr->offset.tag == REG || vstr->offset.tag == VREG){
                oprs.push(vstr->offset);
            }
        } break;
        case MI_BRANCH: {
            auto br_I = (MI_Branch *)I;
            if (br_I->cond != NO_CONDITION) {
                oprs.push(br_I->lhs);
                oprs.push(br_I->rhs);
            }
        } break;
        case MI_VCVT: {
            auto fcvt = (MI_VCvt *)I;
            if(fcvt->from_type == S32) {
                oprs.push(fcvt->src);
            }
        } break;

        case MI_POP:
        case MI_PUSH:
        
        //----float----(return nothing in get_uses())
        case MI_VCMP_SET:
        case MI_VBINARY:
        case MI_VPUSH:
        case MI_VPOP:
        break;
        default: {
            printf("unknown instrID: %d\n", I->tag);
            assert(false && "meet unknown instr in get_uses");
        }
    }

    return oprs;
}

Array<MOperand> s_get_defs(MI *I) {
    Array<MOperand> oprs;
    switch(I->tag) {
        case MI_VMOVE:   { oprs.push(((MI_VMove *)I)->dst); } break;
        case MI_VBINARY: { oprs.push(((MI_VBinary *)I)->dst); } break;

        case MI_VLOAD:   { oprs.push(((MI_VLoad *)I)->reg); } break;

        case MI_VCMP_SET: { oprs.push(((MI_VCmp_Set *)I)->dst); } break;

        case MI_VCVT: {
            auto fcvt = (MI_VCvt *)I;
            if(fcvt->to_type == F32) {
                oprs.push(fcvt->dst);
            }
        } break;

        case MI_FUNC_CALL: {
            for(uint8 r = f0; r < SREG_COUNT; r++) {
                if (s_is_caller_save(r)) {
                    oprs.push(make_sreg(r));
                }
            }
        } break;

        case MI_VPUSH:
        case MI_VPOP:
        case MI_VSTORE: // @TODO: store/load may use self increment thing?
        case MI_MOVE:
        case MI_BINARY:
        case MI_CLZ:
        case MI_RETURN:
        case MI_BRANCH:
        case MI_COMPARE:
        case MI_PUSH:
        case MI_POP:
        case MI_LOAD: // ldr
        case MI_STORE: // str
        case MI_SLT:
        break;
        default: {
            printf("unknown instrID: %d\n", I->tag);
            assert(false && "meet unknown instr in s_get_defs");
        }
    }
    return oprs;
}

Array<MOperand> s_get_uses(MI *I, bool func_has_return_value) {
    Array<MOperand> oprs;

    switch(I->tag) {

        case MI_VMOVE:   {
            if (!(((MI_VMove *)I)->from_s32)) {
                oprs.push(((MI_VMove *)I)->src);
            }
        } break;

        case MI_VBINARY: {
            oprs.push(((MI_VBinary *)I)->lhs);
            if (((MI_VBinary *)I)->fneg == 0) {
                oprs.push(((MI_VBinary *)I)->rhs);
            }
        } break;

        case MI_VCMP_SET: {
            oprs.push(((MI_VCmp_Set *)I)->lhs);
            oprs.push(((MI_VCmp_Set *)I)->rhs);
        } break;

        // 'base' and 'offset' are vreg not vsreg
        case MI_VSTORE: { 
            oprs.push(((MI_VStore *)I)->reg);
        } break;

        case MI_VLOAD: {} break;

        case MI_VCVT: {
            auto fcvt = (MI_VCvt *)I;
            if(fcvt->from_type == F32) {
                oprs.push(fcvt->src);
            }
        } break;

        case MI_FUNC_CALL: {
            for(uint8 r = f10; r < f14; r++) {
                oprs.push(make_sreg(r));
            }
        } break;

        case MI_RETURN: {
            if (func_has_return_value) {
                oprs.push(make_sreg(f10));
            }
        } break;

        case MI_SLT:
        case MI_VPOP:
        case MI_VPUSH:
        case MI_MOVE:
        case MI_BINARY:
        case MI_CLZ:
        case MI_BRANCH:
        case MI_COMPARE:
        case MI_PUSH:
        case MI_POP:
        case MI_LOAD: // ldr
        case MI_STORE: // str
        break;
        default: {
            printf("unknown instrID: %d\n", I->tag);
            assert(false && "meet unknown instr in s_get_uses");
        }

    }

    return oprs;
}