
#include "asm_passes.h"
#include <algorithm>
#include <cmath>

namespace RA
{

    // liveIn and liveOut for each machine blocks
    static Array<Set<MOperand>> live_in;  // new operands in a specific liveness period
    static Array<Set<MOperand>> live_out; // operands used by successor liveness periods

    static Array<Set<MOperand>> use_set; // actually 'use without def' set
    static Array<Set<MOperand>> def_set;

    void analyse_liveness(Func_Asm *func_asm)
    {

        live_in.release();
        live_out.release();
        use_set.release();
        def_set.release();

        live_in.set_len(func_asm->mbs.len);
        live_out.set_len(func_asm->mbs.len);
        use_set.set_len(func_asm->mbs.len);
        def_set.set_len(func_asm->mbs.len);

        // get use and def sets for each machine block
        for (int i = 0; i < func_asm->mbs.len; i++)
        {
            for (auto I = func_asm->mbs[i]->inst; I; I = I->next)
            {
                Array<MOperand> defs = get_defs(I);
                Array<MOperand> uses = get_uses(I, func_asm->has_return_value);

                for (auto use : uses)
                {
                    if (use.tag == VREG && def_set[i].find(use) == def_set[i].end())
                    {
                        use_set[i].insert(use);
                    }
                }

                for (auto def : defs)
                {
                    def_set[i].insert(def);
                }
            }
        }

        // iteratively updating 'live_in' and 'live_out' until stable
        bool changed = true;
        while (changed)
        {
            changed = false;
            for (int i = 0; i < func_asm->mbs.len; i++)
            {
                Set<MOperand> old_in = live_in[i];
                Set<MOperand> old_out = live_out[i];

                live_in[i] = use_set[i];
                for (auto o : live_out[i])
                {
                    if (def_set[i].find(o) == def_set[i].end())
                    {
                        live_in[i].insert(o);
                    }
                }

                live_out[i].clear();
                for (auto succ : func_asm->mbs[i]->succs)
                {
                    for (auto succ_in : live_in[succ->i])
                    {
                        // print_operand(succ_in)
                        live_out[i].insert(succ_in);
                    }
                }

                if (!changed)
                {
                    if ((live_in[i] != old_in) || (live_out[i] != old_out))
                    {
                        changed = true;
                    }
                }
            }
        }
    }

    void s_analyse_liveness(Func_Asm *func_asm)
    {

        live_in.release();
        live_out.release();
        use_set.release();
        def_set.release();

        live_in.set_len(func_asm->mbs.len);
        live_out.set_len(func_asm->mbs.len);
        use_set.set_len(func_asm->mbs.len);
        def_set.set_len(func_asm->mbs.len);

        printf(">>> Begin s_analyse_liveness...\n");
        for (int i = 0; i < func_asm->mbs.len; i++)
        {
            for (auto I = func_asm->mbs[i]->inst; I; I = I->next)
            {
                // printf("s_analyse_liveness InstrID: %d\n", I->tag);
                Array<MOperand> defs = s_get_defs(I);
                Array<MOperand> uses = s_get_uses(I, func_asm->has_return_value);

                // if (defs.len)
                //     printf("sreg defs: ");
                // for (int i = 0; i < defs.len; i++)
                // {
                //     print_operand(defs[i]);
                //     printf(" ");
                // }
                // if (defs.len)
                //     printf("\n");

                // if (uses.len)
                //     printf("sreg uses: ");
                // for (int i = 0; i < uses.len; i++)
                // {
                //     print_operand(uses[i]);
                //     printf(" ");
                // }
                // if (uses.len)
                //     printf("\n");

                for (auto use : uses)
                {
                    if (use.tag == VSREG && def_set[i].find(use) == def_set[i].end())
                    {
                        use_set[i].insert(use);
                    }
                }

                for (auto def : defs)
                {
                    def_set[i].insert(def);
                }
            }
        }

        bool changed = true;
        while (changed)
        {
            changed = false;
            for (int i = 0; i < func_asm->mbs.len; i++)
            {
                Set<MOperand> old_in = live_in[i];
                Set<MOperand> old_out = live_out[i];

                // live_in[i] = use_set[i] union (live_out[i]-def_set[i])
                live_in[i] = use_set[i];
                for (auto o : live_out[i])
                {
                    if (def_set[i].find(o) == def_set[i].end())
                    {
                        live_in[i].insert(o);
                    }
                }

                live_out[i].clear();
                for (auto succ : func_asm->mbs[i]->succs)
                {
                    for (auto succ_in : live_in[succ->i])
                    {
                        // print_operand(succ_in)
                        live_out[i].insert(succ_in);
                    }
                }

                if (!changed)
                {
                    if ((live_in[i] != old_in) || (live_out[i] != old_out))
                    {
                        changed = true;
                    }
                }
            }
        }
        printf(">>> End s_analyse_liveness...\n");
    }
    // @TODO: not certain
    const int K = 21; // t0~t7, s0~s11, al
    const int s_K = 32;
    Set<MOperand> precolored;
    Set<MOperand> initial;
    Set<MOperand> simplify_worklist;
    Set<MOperand> freeze_worklist;
    Set<MOperand> spill_worklist;
    Set<MOperand> spilled_nodes;
    Set<MOperand> coalesced_nodes;
    Set<MOperand> colored_nodes;
    Set<MOperand> select_set;
    Array<MOperand> select_stack;
    Array<uint8> needs_save;

    // keep track of already spilled nodes
    // avoid spilling repeatly
    Set<MOperand> already_spilled;

    Set<MI_Move *, MI_Move_Pointer_Cmp> coalesced_moves;
    Set<MI_Move *, MI_Move_Pointer_Cmp> constrained_moves;
    Set<MI_Move *, MI_Move_Pointer_Cmp> frozen_moves;
    Set<MI_Move *, MI_Move_Pointer_Cmp> worklist_moves;
    Set<MI_Move *, MI_Move_Pointer_Cmp> active_moves;

    Set<MI_VMove *, MI_VMove_Pointer_Cmp> s_coalesced_moves;
    Set<MI_VMove *, MI_VMove_Pointer_Cmp> s_constrained_moves;
    Set<MI_VMove *, MI_VMove_Pointer_Cmp> s_frozen_moves;
    Set<MI_VMove *, MI_VMove_Pointer_Cmp> s_worklist_moves;
    Set<MI_VMove *, MI_VMove_Pointer_Cmp> s_active_moves;

    typedef Pair<MOperand, MOperand> Edge;
    Set<Edge> adj_set;
    // Set< Set<MOperand> > adj_list;
    Map<MOperand, Set<MOperand>> adj_list;
    Map<MOperand, int> degree; // @default 0?

    Map<MOperand, Set<MI_Move *, MI_Move_Pointer_Cmp>> move_list;
    Map<MOperand, Set<MI_VMove *, MI_VMove_Pointer_Cmp>> s_move_list;
    Map<MOperand, MOperand> alias;
    Map<MOperand, int32> color;
    Map<MOperand, int32> s_color;
    Map<MOperand, uint32> use_def_count;
    Map<MOperand, uint32> loop_depth;

    int32 spilled_stack_size = 0;
    bool done_post_work = false;
    bool need_to_legalize_imm = false;

#define in(it, set) (((set).find(it)) != ((set).end()))

    bool need_alloc(MOperand opr)
    {
        return (opr.tag == REG || opr.tag == VREG);
    }

    bool s_need_alloc(MOperand opr)
    {
        return (opr.tag == SREG || opr.tag == VSREG);
    }

    void add_edge(MOperand u, MOperand v);
    void build(Func_Asm *func_asm);
    void s_build(Func_Asm *func_asm);
    Set<MOperand> adjacent(MOperand n);
    // Set<MI_Move *> node_moves(MOperand n);
    bool move_related(MOperand n);
    bool s_move_related(MOperand n);
    void make_worklist();
    void s_make_worklist();
    void enable_moves(Set<MOperand> nodes);
    void s_enable_moves(Set<MOperand> nodes);
    void decrement_degree(MOperand m);
    void s_decrement_degree(MOperand m);
    void simplify();
    void coalesce();
    void add_worklist(MOperand u);
    void s_simplify();
    void s_coalesce();
    void s_add_worklist(MOperand u);
    bool ok(MOperand t, MOperand r);
    bool s_ok(MOperand t, MOperand r);
    bool conservative(Set<MOperand> nodes);
    bool s_conservative(Set<MOperand> nodes);
    MOperand get_alias(MOperand n);
    void combine(MOperand u, MOperand v);
    void freeze();
    void freeze_moves(MOperand u);
    void s_combine(MOperand u, MOperand v);
    void s_freeze();
    void s_freeze_moves(MOperand u);
    void assign_colors();
    void s_assign_colors();
    void select_spill();
    void s_select_spill();

    void add_edge(MOperand u, MOperand v)
    {
        auto p = Edge{u, v};
        if (!in(p, adj_set) && !(u == v))
        {
            // printf("interference: {"); print_operand(u); printf(", "); print_operand(v); printf("}\n");
            adj_set.insert({u, v});
            adj_set.insert({v, u});

            if (u.tag != REG)
            {
                adj_list[u].insert(v);
                degree[u]++;
            }

            if (v.tag != REG)
            {
                adj_list[v].insert(u);
                degree[v]++;
            }
        }
    }

    void s_add_edge(MOperand u, MOperand v)
    {
        auto p = Edge{u, v};
        if (!in(p, adj_set) && !(u == v))
        {
            // printf("interference: {");
            // print_operand(u);
            // printf(", ");
            // print_operand(v);
            // printf("}\n");
            adj_set.insert({u, v});
            adj_set.insert({v, u});

            if (u.tag != SREG)
            {
                adj_list[u].insert(v);
                degree[u]++;
            }

            if (v.tag != SREG)
            {
                adj_list[v].insert(u);
                degree[v]++;
            }
        }
    }

    void build(Func_Asm *func_asm)
    {
        int cnt = 0;
        for (auto b : func_asm->mbs)
        {
            int idx = b->i;
            auto live = live_out[b->i];
            cout << "Basic block " << cnt++ << endl;
            for (auto I = func_asm->mbs[b->i]->last_inst; I; I = I->prev)
            {
                auto defs = get_defs(I);
                auto uses = get_uses(I, func_asm->has_return_value);
                if (I->tag == MI_MOVE)
                {
                    if (need_alloc(defs[0]) && need_alloc(uses[0]))
                    {
                        for (auto use : uses)
                        {
                            if (!need_alloc(use))
                                continue;
                            live.erase(use);
                            move_list[use].insert((MI_Move *)I);
                        }

                        for (auto def : defs)
                        {
                            move_list[def].insert((MI_Move *)I);
                        }

                        worklist_moves.insert((MI_Move *)I);
                    }
                }

                // @Bug, needs to be allocated?
                for (auto def : defs)
                {
                    if (need_alloc(def))
                    {
                        live.insert(def);
                        use_def_count[def]++;
                        if (b->loop_depth > loop_depth[def])
                        {
                            loop_depth[def] = b->loop_depth;
                        }
                        for (auto l : live)
                            add_edge(l, def);
                    }
                }

                // @Optimize
                Set<MOperand> new_live;
                for (auto u : uses)
                {
                    if (need_alloc(u))
                    {
                        use_def_count[u]++;
                        if (b->loop_depth > loop_depth[u])
                        {
                            loop_depth[u] = b->loop_depth;
                        }
                        new_live.insert(u);
                    }
                }
                for (auto l : live)
                {
                    // if (need_alloc(def) && l != def) {
                    if (defs.find(l) == -1)
                    {
                        new_live.insert(l);
                    }
                }
                live = new_live;
            }
        }
    }

    void s_build(Func_Asm *func_asm)
    {
        int cnt = 0;
        for (auto b : func_asm->mbs)
        {
            int idx = b->i;
            auto live = live_out[b->i];
            cout << "Basic block " << cnt++ << endl;
            for (auto I = func_asm->mbs[b->i]->last_inst; I; I = I->prev)
            {
                auto defs = s_get_defs(I);
                auto uses = s_get_uses(I, func_asm->has_return_value);
                if (I->tag == MI_VMOVE && !(((MI_VMove *)I)->from_s32))
                {
                    if (s_need_alloc(defs[0]) && s_need_alloc(uses[0]))
                    {
                        for (auto use : uses)
                        {
                            if (!s_need_alloc(use))
                                continue;
                            live.erase(use);
                            s_move_list[use].insert((MI_VMove *)I);
                        }

                        for (auto def : defs)
                        {
                            s_move_list[def].insert((MI_VMove *)I);
                        }

                        s_worklist_moves.insert((MI_VMove *)I);
                    }
                }

                // @Bug, needs to be allocated?
                for (auto def : defs)
                {
                    if (s_need_alloc(def))
                    {
                        live.insert(def);
                        use_def_count[def]++;
                        if (b->loop_depth > loop_depth[def])
                        {
                            loop_depth[def] = b->loop_depth;
                        }
                        for (auto l : live)
                            s_add_edge(l, def);
                    }
                }

                // @Optimize
                Set<MOperand> new_live;
                for (auto u : uses)
                {
                    if (s_need_alloc(u))
                    {
                        use_def_count[u]++;
                        if (b->loop_depth > loop_depth[u])
                        {
                            loop_depth[u] = b->loop_depth;
                        }
                        new_live.insert(u);
                    }
                }
                for (auto l : live)
                {
                    // if (need_alloc(def) && l != def) {
                    if (defs.find(l) == -1)
                    {
                        new_live.insert(l);
                    }
                }
                live = new_live;
            }
        }
    }
    Set<MOperand> adjacent(MOperand n)
    {
        Set<MOperand> a;
        auto it = adj_list[n].begin();
        while (it != adj_list[n].end())
        {
            if (!in(*it, select_set) && !in(*it, coalesced_nodes))
            {
                a.insert(*it);
            }
            it++;
        }
        return a;
    }

    Set<MI_Move *> node_moves(MOperand n)
    {
        Set<MI_Move *> s;
        auto it = move_list[n].begin();
        while (it != move_list[n].end())
        {
            if (in(*it, active_moves) || in(*it, worklist_moves))
            {
                s.insert(*it);
            }
            it++;
        }
        return s;
    }

    Set<MI_VMove *> s_node_moves(MOperand n)
    {
        Set<MI_VMove *> s;
        auto it = s_move_list[n].begin();
        while (it != s_move_list[n].end())
        {
            if (in(*it, s_active_moves) || in(*it, s_worklist_moves))
            {
                s.insert(*it);
            }
            it++;
        }
        return s;
    }

    bool move_related(MOperand n)
    {
        return !node_moves(n).empty();
    }

    bool s_move_related(MOperand n)
    {
        return !s_node_moves(n).empty();
    }

    void make_worklist()
    {
        while (!initial.empty())
        {
            auto n = *(initial.begin());
            initial.erase(n);
            if (degree[n] >= K)
            {
                spill_worklist.insert(n);
            }
            else if (move_related(n))
            {
                freeze_worklist.insert(n);
            }
            else
            {
                simplify_worklist.insert(n);
            }
        }
    }

    void s_make_worklist()
    {
        while (!initial.empty())
        {
            auto n = *(initial.begin());
            initial.erase(n);
            if (degree[n] >= s_K)
            {
                spill_worklist.insert(n);
            }
            else if (s_move_related(n))
            {
                freeze_worklist.insert(n);
            }
            else
            {
                simplify_worklist.insert(n);
            }
        }
    }

    void enable_moves(Set<MOperand> nodes)
    {
        for (auto n : nodes)
        {
            for (auto m : node_moves(n))
            {
                if (in(m, active_moves))
                {
                    active_moves.erase(m);
                    worklist_moves.insert(m);
                }
            }
        }
    }

    void s_enable_moves(Set<MOperand> nodes)
    {
        for (auto n : nodes)
        {
            for (auto m : s_node_moves(n))
            {
                if (in(m, s_active_moves))
                {
                    s_active_moves.erase(m);
                    s_worklist_moves.insert(m);
                }
            }
        }
    }

    void decrement_degree(MOperand m)
    {
        int d = degree[m]--;
        if (d == K)
        {
            auto adj = adjacent(m);
            adj.insert(m);
            enable_moves(adj);
            spill_worklist.erase(m);
            if (move_related(m))
            {
                freeze_worklist.insert(m);
            }
            else
            {
                simplify_worklist.insert(m);
            }
        }
    }

    void s_decrement_degree(MOperand m)
    {
        int d = degree[m]--;
        if (d == s_K)
        {
            auto adj = adjacent(m);
            adj.insert(m);
            s_enable_moves(adj);
            spill_worklist.erase(m);
            if (s_move_related(m))
            {
                freeze_worklist.insert(m);
            }
            else
            {
                simplify_worklist.insert(m);
            }
        }
    }

    void simplify()
    {
        auto n = *(simplify_worklist.begin());
        simplify_worklist.erase(n);
        printf("simplifying ");
        print_operand(n);
        printf("\n");
        select_stack.push(n);
        select_set.insert(n);
        for (auto m : adjacent(n))
        {
            decrement_degree(m);
        }
    }

    void s_simplify()
    {
        auto n = *(simplify_worklist.begin());
        simplify_worklist.erase(n);
        printf("simplifying ");
        print_operand(n);
        printf("\n");
        select_stack.push(n);
        select_set.insert(n);
        for (auto m : adjacent(n))
        {
            s_decrement_degree(m);
        }
    }

    // delete 'mv dst, src' and replace 'dst' with 'src'
    void coalesce()
    {
        auto m = *(worklist_moves.begin());
        worklist_moves.erase(m);
        printf("coalescing ");
        print_operand(m->dst);
        printf(" <- ");
        print_operand(m->src);
        printf("\n");
        auto u = get_alias(m->src);
        auto v = get_alias(m->dst);
        auto e = Edge{u, v};
        if (v.tag == REG)
        { // @What?
            auto t = u;
            u = v;
            v = t;
        }

        // @Performance
        int okok = false;
        for (auto t : adjacent(v))
        {
            if (ok(t, u))
            {
                okok = true;
                break;
            }
        }

        auto join = adjacent(u);
        for (auto n : adjacent(v))
        {
            join.insert(n);
        }

        if (u == v)
        {
            coalesced_moves.insert(m);
            add_worklist(u);
        }
        else if (v.tag == REG || in(e, adj_set))
        {
            constrained_moves.insert(m);
            add_worklist(u);
            add_worklist(v);
        }
        else if ((u.tag == REG && okok) ||
                 (u.tag != REG && conservative(join)))
        {
            coalesced_moves.insert(m);
            combine(u, v);
            add_worklist(u);
        }
        else
        {
            active_moves.insert(m);
        }
    }

    void s_coalesce()
    {
        auto m = *(s_worklist_moves.begin());
        s_worklist_moves.erase(m);
        printf("coalescing ");
        print_operand(m->dst);
        printf(" <- ");
        print_operand(m->src);
        printf("\n");
        auto u = get_alias(m->src);
        auto v = get_alias(m->dst);
        auto e = Edge{u, v};
        if (v.tag == SREG)
        { // @What?
            auto t = u;
            u = v;
            v = t;
        }

        // @Performance
        int okok = false;
        for (auto t : adjacent(v))
        {
            if (s_ok(t, u))
            {
                okok = true;
                break;
            }
        }

        auto join = adjacent(u);
        for (auto n : adjacent(v))
        {
            join.insert(n);
        }

        if (u == v)
        {
            s_coalesced_moves.insert(m);
            s_add_worklist(u);
        }
        else if (v.tag == SREG || in(e, adj_set))
        {
            s_constrained_moves.insert(m);
            s_add_worklist(u);
            s_add_worklist(v);
        }
        else if ((u.tag == SREG && okok) ||
                 (u.tag != SREG && s_conservative(join)))
        {
            s_coalesced_moves.insert(m);
            s_combine(u, v);
            s_add_worklist(u);
        }
        else
        {
            s_active_moves.insert(m);
        }
    }

    void add_worklist(MOperand u)
    {
        if (u.tag != REG && !move_related(u) && degree[u] < K)
        {
            freeze_worklist.erase(u);
            simplify_worklist.insert(u);
        }
    }

    void s_add_worklist(MOperand u)
    {
        if (u.tag != SREG && !s_move_related(u) && degree[u] < s_K)
        {
            freeze_worklist.erase(u);
            simplify_worklist.insert(u);
        }
    }

    bool ok(MOperand t, MOperand r)
    {
        auto p = Edge{t, r};
        return ((degree[t] < K) || (t.tag == REG) || (in(p, adj_set)));
    }

    bool s_ok(MOperand t, MOperand r)
    {
        auto p = Edge{t, r};
        return ((degree[t] < s_K) || (t.tag == SREG) || (in(p, adj_set)));
    }

    bool conservative(Set<MOperand> nodes)
    {
        int k = 0;
        for (auto n : nodes)
        {
            if (degree[n] >= K)
            {
                k++;
            }
        }
        return k < K;
    }

    bool s_conservative(Set<MOperand> nodes)
    {
        int k = 0;
        for (auto n : nodes)
        {
            if (degree[n] >= s_K)
            {
                k++;
            }
        }
        return k < s_K;
    }

    MOperand get_alias(MOperand n)
    {
        if (in(n, coalesced_nodes))
        {
            return get_alias(alias[n]);
        }
        else
        {
            return n;
        }
    }

    // use 'u' to replace 'v'
    void combine(MOperand u, MOperand v)
    {
        if (in(v, freeze_worklist))
        {
            freeze_worklist.erase(v);
        }
        else
        {
            spill_worklist.erase(v);
        }
        coalesced_nodes.insert(v);
        alias[v] = u;

        // FIXME
        for (auto n : move_list[v])
        {
            move_list[u].insert(n);
        }

        for (auto t : adjacent(v))
        {
            add_edge(t, u);
            decrement_degree(t);
        }

        if (degree[u] >= K && in(u, freeze_worklist))
        {
            freeze_worklist.erase(u);
            spill_worklist.insert(u);
        }
    }

    void s_combine(MOperand u, MOperand v)
    {
        if (in(v, freeze_worklist))
        {
            freeze_worklist.erase(v);
        }
        else
        {
            spill_worklist.erase(v);
        }
        coalesced_nodes.insert(v);
        alias[v] = u;

        // FIXME
        for (auto n : s_move_list[v])
        {
            s_move_list[u].insert(n);
        }

        for (auto t : adjacent(v))
        {
            s_add_edge(t, u);
            s_decrement_degree(t);
        }

        if (degree[u] >= s_K && in(u, freeze_worklist))
        {
            freeze_worklist.erase(u);
            spill_worklist.insert(u);
        }
    }

    void freeze()
    {
        auto u = *(freeze_worklist.begin());
        freeze_worklist.erase(u);
        simplify_worklist.insert(u);
        freeze_moves(u);
    }

    void s_freeze()
    {
        auto u = *(freeze_worklist.begin());
        freeze_worklist.erase(u);
        simplify_worklist.insert(u);
        s_freeze_moves(u);
    }

    void freeze_moves(MOperand u)
    {
        for (auto m : node_moves(u))
        {
            // @Order?
            auto u = m->src;
            auto v = m->dst;

            if (in(m, active_moves))
            {
                active_moves.erase(m);
            }
            else
            {
                worklist_moves.erase(m);
            }

            frozen_moves.insert(m);
            if (node_moves(v).empty() && degree[v] < K)
            {
                freeze_worklist.erase(v);
                simplify_worklist.insert(v);
            }
        }
    }

    void s_freeze_moves(MOperand u)
    {
        for (auto m : s_node_moves(u))
        {
            // @Order?
            auto u = m->src;
            auto v = m->dst;

            if (in(m, s_active_moves))
            {
                s_active_moves.erase(m);
            }
            else
            {
                s_worklist_moves.erase(m);
            }

            s_frozen_moves.insert(m);
            if (s_node_moves(v).empty() && degree[v] < s_K)
            {
                freeze_worklist.erase(v);
                simplify_worklist.insert(v);
            }
        }
    }

    // @Optimization: can we use lr?
    // temporary regs and saved regs, can we use al???
    bool is_general_purpose_register(uint8 r)
    {
        return ((r >= x5) && (r <= x9)) || ((r >= x18) && (r <= x31)) || r == ra;
        // return ((r >= r0) && (r <= r12)) || (r == lr);
        // return ((r >= r0) && (r <= r11));
    }

    void assign_colors()
    {
        while (!select_stack.empty())
        {
            auto n = select_stack.back();
            select_stack.pop();
            Array<uint8> ok_colors;

            // assign caller save first
            for (uint8 r = x10; r <= x17; r++)
            { // a0~a7
                ok_colors.push(r);
            }
            ok_colors.push(x5); // t0~t2
            ok_colors.push(x6);
            ok_colors.push(x7);
            for (uint8 r = x28; r <= x31; r++)
            { // t3~t6
                ok_colors.push(r);
            }

            // assign callee save regs
            ok_colors.push(x9);
            for (uint8 r = x18; r <= x27; r++)
            { // s2~s11
                ok_colors.push(r);
            }
            ok_colors.push(x8); // s0(fp)
            // @TODO: may incur bug?
            ok_colors.push(ra);

            // remove adjcent colors(heuristic algorithm)
            for (auto w : adj_list[n])
            {
                auto a = get_alias(w);
                if (in(a, colored_nodes) || in(a, precolored))
                {
                    ok_colors.remove(color[a]);
                }
            }
            if (ok_colors.empty())
            {
                spilled_nodes.insert(n);
            }
            else
            {
                colored_nodes.insert(n);
                auto c = ok_colors.front();
                color[n] = c;
            }
        }
        for (auto n : coalesced_nodes)
        {
            color[n] = color[get_alias(n)];
        }
        cout << "my corloring result:" << endl;
        for (auto it : color)
        {
            cout << it.first.adr << ": " << it.second << endl;
            // assert(it.second != zero);
        }
    }

    void select_spill()
    {
        printf("spilled!\n");
        // auto m = *(spill_worklist.begin());
        MOperand m = {};
        double min_cost = 1e40;
        bool is_already_spilled = false;
        Set<MOperand> spilled;
        for (auto it = spill_worklist.begin();
             it != spill_worklist.end();
             it++)
        {
            // when regs already spilled, erase from spill_worklist
            if (in(*it, already_spilled)) {
                is_already_spilled = true;
                spilled.insert(*it);
                continue;
            }

            assert(degree[*it] != 0);

            double cost = use_def_count[*it] * pow(10, loop_depth[*it]) / degree[*it];
            // print_operand(*it);
            // printf(": use_def_count: %d, loop_depth: %d, degree: %d, cost: %f\n",
            //        use_def_count[*it], loop_depth[*it], degree[*it], cost);

            if (cost < min_cost)
            {
                m = *it;
                min_cost = cost;
            }
        }

        if (is_already_spilled) {
            for (auto it = spilled.begin(); it != spilled.end(); it++) {
                spill_worklist.erase(*it);
            }
            if (m.tag == ERRORTYPE) return;
        }
        assert(m.tag != ERRORTYPE);

        spill_worklist.erase(m);
        simplify_worklist.insert(m);
        freeze_moves(m);
    }

    void s_select_spill()
    {
        printf("spilled!\n");
        // auto m = *(spill_worklist.begin());
        MOperand m = {};
        double min_cost = 1e40;
        bool is_already_spilled = false;
        Set<MOperand> spilled;
        for (auto it = spill_worklist.begin();
             it != spill_worklist.end();
             it++)
        {

            // when regs already spilled, erase from spill_worklist
            if (in(*it, already_spilled)) {
                is_already_spilled = true;
                spilled.insert(*it);
                continue;
            }

            assert(degree[*it] != 0);

            double cost = use_def_count[*it] * pow(10, loop_depth[*it]) / degree[*it];
            // print_operand(*it);
            // printf(": use_def_count: %d, loop_depth: %d, degree: %d, cost: %f\n",
            //        use_def_count[*it], loop_depth[*it], degree[*it], cost);

            if (cost < min_cost)
            {
                m = *it;
                min_cost = cost;
            }
        }

        if (is_already_spilled) {
            for (auto it = spilled.begin(); it != spilled.end(); it++) {
                spill_worklist.erase(*it);
            }
            if (m.tag == ERRORTYPE) return;
        }

        assert(m.tag != ERRORTYPE);

        spill_worklist.erase(m);
        simplify_worklist.insert(m);
        s_freeze_moves(m);
    }

    void color_register(MOperand &m)
    {
        if (need_alloc(m))
        {
            uint8 reg = color[m];
            m.value = reg;
            if (m.tag == VREG)
                m.tag = REG;
        }

        if (m.tag == REG && is_callee_save(m.value))
        {
            if (needs_save.find(m.value) == -1)
            {
                needs_save.push(m.value);
            }
        }
    }

    void s_color_register(MOperand &m)
    {
        if (s_need_alloc(m))
        {
            uint8 reg = s_color[m];
            m.value = reg;
            if (m.tag == VSREG)
                m.tag = SREG;
        }

        if (m.tag == SREG && is_callee_save_sreg(m.value))
        {
            if (needs_save.find(m.value) == -1)
            {
                needs_save.push(m.value);
            }
        }
    }

    void allocate_register(Func_Asm *func_asm)
    {
        precolored.clear();
        initial.clear();
        simplify_worklist.clear();
        freeze_worklist.clear();
        spill_worklist.clear();
        spilled_nodes.clear();
        coalesced_nodes.clear();
        colored_nodes.clear();
        coalesced_moves.clear();
        s_coalesced_moves.clear();
        constrained_moves.clear();
        s_constrained_moves.clear();
        frozen_moves.clear();
        s_frozen_moves.clear();
        select_set.clear();
        worklist_moves.clear();
        s_worklist_moves.clear();
        active_moves.clear();
        s_active_moves.clear();
        adj_set.clear();
        adj_list.clear();
        degree.clear();
        move_list.clear();
        s_move_list.clear();
        alias.clear();
        color.clear();
        s_color.clear();
        use_def_count.clear();
        loop_depth.clear();

        select_stack.len = 0;
        needs_save.len = 0;

        printf(">>> Analysing liveness of function %s\n", func_asm->name.c_str());
        analyse_liveness(func_asm);
        printf(">>> Finished analysing liveness of function %s...\n\n", func_asm->name.c_str());

        // @Check: precolored register?
        // x0~x4 are specific regs, x0 can not be written
        for (int i = 1; i < REG_COUNT; i++)
        {
            auto r = make_reg(i);
            precolored.insert(r);
            color[r] = i;
        }

        for (int i = 0; i < func_asm->vreg_count; i++)
        {
            initial.insert(make_vreg(i));
        }

        printf(">>> Building interference graph for coloring...\n");
        build(func_asm);
        printf(">>> done building\n");
        make_worklist();
        printf(">>> worklist done\n");
        do
        {
            if (!simplify_worklist.empty())
                simplify();
            else if (!worklist_moves.empty())
                coalesce();
            else if (!freeze_worklist.empty())
                freeze();
            else if (!spill_worklist.empty())
                select_spill();
        } while (
            !simplify_worklist.empty() ||
            !worklist_moves.empty() ||
            !freeze_worklist.empty() ||
            !spill_worklist.empty());

        printf("assigning\n");
        assign_colors();
        printf("assigned\n");

        printf("\n");
        if (!spilled_nodes.empty())
        {
            // rewrite_program(spilled_nodes);

            // rewrite program
            // insert stores and restores for each spilled node
            for (MOperand n : spilled_nodes)
            {

                already_spilled.insert(n);

                printf("spilling nodes: ");
                print_operand(n);
                printf("\n");
                spilled_stack_size += 8; // Alignment

                for (auto bb : func_asm->mbs)
                {
                    for (auto I = bb->inst; I; I = I->next)
                    {
                        auto defs = get_defs(I);
                        auto uses = get_uses(I, func_asm->has_return_value);

                        bool inserted_store = false;
                        bool inserted_use = false;
                        // insert store after def
                        if (defs.find(n) != -1)
                        {
                            auto str = new MI_Store;
                            str->mem_tag = MEM_SAVE_SPILL;
                            str->reg = make_vreg(func_asm->vreg_count++);
                            str->base = make_reg(sp);
                            str->offset = make_imm(-(spilled_stack_size));
                            str->is_rv64 = true;
                            already_spilled.insert(str->reg);

                            replace_defs(I, n, str->reg);
                            insert((MI *)str, I->next); // @Last?
                            I = I->next;                // jump past newly inserted str
                            inserted_store = true;
                        }

                        // insert restore before use
                        if (uses.find(n) != -1)
                        {
                            auto ldr = new MI_Load;
                            ldr->mem_tag = MEM_LOAD_SPILL;
                            ldr->reg = make_vreg(func_asm->vreg_count++);
                            ldr->base = make_reg(sp);
                            ldr->offset = make_imm(-(spilled_stack_size));
                            ldr->is_rv64 = true;
                            already_spilled.insert(ldr->reg);
                            replace_uses(I, n, ldr->reg);
                            // @TODO: get uses based on arg count
                            insert((MI *)ldr, I);
                            inserted_use = true;
                        }
                        assert(!(inserted_use && inserted_store));
                        if (inserted_use && inserted_store)
                        {
                            exit(64);
                        }
                    }
                    // bb->erase_marked_values();
                }
            }

            printf(">>> spilled stack size: %d\n", spilled_stack_size);
            printf(">>> after spilling: \n");
            print_function_asm(func_asm);
            printf("\n");

            allocate_register(func_asm);
        }
        else
        {

            // @HACK
            // we need to know needs_save to setup the stack
            // which is computed by color_register
            // which runs after we setup the stack...
            for (auto bb : func_asm->mbs)
            {
                for (auto I = bb->inst; I; I = I->next)
                {
                    for (auto d : get_defs(I))
                    {
                        color_register(d);
                    }
                }
            }

            if (!done_post_work)
            {
                done_post_work = true;

                func_asm->stack_size += spilled_stack_size;
                func_asm->reg_needs_save = needs_save.size();
                func_asm->reg_spill_size += spilled_stack_size;

                printf(">>> registers that need to save: ");

                for (auto r : needs_save)
                {
                    printf("x%d ", r);
                }
                printf("\n");

                // allocate stack space for argument passing
                // maximal number of form arguments
                int max_arg_count = 0;
                int stack_arg_size = 0;
                for (auto mb : func_asm->mbs)
                {
                    for (auto I = mb->inst; I; I = I->next)
                    {
                        if (I->tag == MI_FUNC_CALL)
                        {
                            auto call = (MI_Func_Call *)I;
                            if (call->arg_count > max_arg_count)
                            {
                                max_arg_count = call->arg_count;
                            }
                            if (call->arg_stack_size > stack_arg_size)
                            {
                                stack_arg_size = call->arg_stack_size;
                            }
                        }
                    }
                }

                if (max_arg_count > 4)
                {
                    // func_asm->stack_size += (max_arg_count - 4)*4;
                    func_asm->stack_size += stack_arg_size;
                }

                // cout << ">>> " << func_asm->name << "  " << func_asm->stack_size << endl;

                // insert add/sub to set up or destroy the stack
                // if (func_asm->stack_size != 0)
                // {
                //     auto alloc_stack = new MI_Binary(BINARY_SUBTRACT, make_reg(sp), make_reg(sp), make_imm(func_asm->stack_size));
                //     alloc_stack->is_rv64 = true;

                //     auto start_inst = func_asm->mbs[0]->inst;
                //     insert((MI *)alloc_stack, start_inst);

                //     for (auto bb : func_asm->mbs)
                //     {
                //         if (bb->last_inst && bb->last_inst->tag == MI_RETURN)
                //         {
                //             auto destroy_stack = new MI_Binary(BINARY_ADD, make_reg(sp), make_reg(sp), make_imm(func_asm->stack_size));
                //             destroy_stack->is_rv64 = true;
                //             insert((MI *)destroy_stack, bb->last_inst);
                //         }
                //     }
                // }

                // convert fp to sp
                // now func_asm->stack_size = size of (local arrays + spilled vars + max calling arguments);
                // not including the callee-save registers' space

                // fixup sp when calculating base address for local arrays
                // for (auto base : func_asm->local_array_bases)
                // {
                //     assert(base->tag == MI_BINARY);
                //     auto sub = (MI_Binary *)base;
                //     assert(sub->op == BINARY_SUBTRACT);
                //     assert(sub->lhs == make_reg(sp));
                //     assert(sub->rhs.tag == IMM);

                //     int32 offset_relative_to_sp = func_asm->stack_size - sub->rhs.value;
                //     // @TODO replace with a mov directly
                //     // if offset relative to sp is 0

                //     sub->op = BINARY_ADD;
                //     sub->lhs = make_reg(sp);
                //     sub->rhs = make_imm(offset_relative_to_sp);
                //     sub->lhs.size = 8;
                //     sub->is_rv64 = true;
                // }

                // for (auto bb : func_asm->mbs)
                // {
                //     for (auto I = bb->inst; I; I = I->next)
                //     {

                //         int save_stack_size = needs_save.len * 8;
                //         // int save_stack_size = 0;
                //         if (I->tag == MI_LOAD)
                //         {
                //             auto ldr = (MI_Load *)I;
                //             if (ldr->mem_tag == MEM_LOAD_SPILL)
                //             {
                //                 int32 offset_value = ((ldr->offset.tag == ERRORTYPE) ? 0 : ldr->offset.value);
                //                 int32 offset_relative_to_sp = func_asm->stack_size + offset_value;
                //                 ldr->base.value = sp;
                //                 ldr->offset = make_imm(offset_relative_to_sp);
                //             }
                //             else if (ldr->mem_tag == MEM_LOAD_ARG)
                //             {
                //                 int32 offset_value = ((ldr->offset.tag == ERRORTYPE) ? 0 : ldr->offset.value);
                //                 int32 offset_relative_to_sp = func_asm->stack_size + offset_value + save_stack_size;
                //                 ldr->base.value = sp;
                //                 ldr->offset = make_imm(offset_relative_to_sp);
                //             }
                //         }

                //         if (I->tag == MI_STORE)
                //         {
                //             auto str = (MI_Store *)I;
                //             if (str->mem_tag == MEM_SAVE_SPILL)
                //             {
                //                 int32 offset_value = ((str->offset.tag == ERRORTYPE) ? 0 : str->offset.value);
                //                 int32 offset_relative_to_sp = func_asm->stack_size + offset_value;
                //                 str->base.value = sp;
                //                 str->offset = make_imm(offset_relative_to_sp);
                //             }
                //         }
                //     }
                // }

                // insert push/pops to save these registers
                if (!needs_save.empty())
                {
                    // insert push in the start block
                    std::sort(needs_save.begin(), needs_save.end());
                    auto store = new MI_Push;
                    for (auto r : needs_save)
                    {
                        store->operands.push(make_reg(r));
                    }
                    insert((MI *)store, func_asm->mbs[0]->inst);

                    // restore registers at exit

                    for (auto bb : func_asm->mbs)
                    {
                        if (bb->last_inst && bb->last_inst->tag == MI_RETURN)
                        {
                            auto restore = new MI_Pop;
                            for (auto r : needs_save)
                            {
                                restore->operands.push(make_reg(r));
                            }
                            insert((MI *)restore, bb->last_inst);
                        }
                    }
                }

                // check imm
                // bool need_to_legalize_imm = false;
                // for (auto mb : func_asm->mbs)
                // {
                //     for (auto I = mb->inst; I; I = I->next)
                //     {
                //         auto uses = get_uses(I, func_asm->has_return_value);

                //         bool inst_need_legalize = false;
                //         MOperand use_of_imm;

                //         if (I->tag == MI_LOAD)
                //         {
                //             auto load = (MI_Load *)I;
                //             if (load->base.tag == IMM)
                //             {
                //                 goto done;
                //             }
                //         }

                //         if (I->tag == MI_LOAD || I->tag == MI_STORE)
                //         {
                //             auto load_or_store = (MI_Load *)I;
                //             if (load_or_store->offset.tag == IMM)
                //             {
                //                 use_of_imm = load_or_store->offset;
                //                 inst_need_legalize = !can_be_imm12(use_of_imm.value);
                //                 goto done;
                //             }
                //         }

                //         if (I->tag == MI_BINARY)
                //         {
                //             auto bi = (MI_Binary *)I;
                //             if (bi->rhs.tag == IMM)
                //             {
                //                 use_of_imm = bi->rhs;
                //                 inst_need_legalize = !can_be_imm12(use_of_imm.value);
                //                 goto done;
                //             }
                //         }

                //     done:;

                //         need_to_legalize_imm |= inst_need_legalize;

                //         if (inst_need_legalize)
                //         {
                //             printf("%d cannot be imm!\n", use_of_imm.value);

                //             // allocate a temp vreg to load the constant from literal pool
                //             auto temp = make_vreg(func_asm->vreg_count++);
                //             MI_Load *ldr = emit_load_of_constant(temp, use_of_imm.value);
                //             insert(ldr, I);

                //             // RISC-V do not allow 'sw a1,a0(s2)'('a0'=temp)
                //             // we should split it into 'add' and 'sw'
                //             if (I->tag == MI_LOAD || I->tag == MI_STORE)
                //             {
                //                 auto load_or_store = (MI_Load *)I;
                //                 auto new_base = make_vreg(func_asm->vreg_count++);
                //                 MI_Binary *bi_add = new MI_Binary(BINARY_ADD, new_base, load_or_store->base, temp);
                //                 bi_add->is_rv64 = true;
                //                 insert(bi_add, I);
                //                 replace_uses(load_or_store, use_of_imm, make_imm(0));
                //                 replace_uses(load_or_store, load_or_store->base, new_base);
                //             }
                //             else
                //             {
                //                 replace_uses(I, use_of_imm, temp);
                //             }
                //         }
                //     }
                // }

                // if (need_to_legalize_imm)
                // {
                //     printf(">>> need to legalize imm!\n");
                //     print_function_asm(func_asm);

                //     int32 last_size = spilled_stack_size;
                //     Array<uint8> last_needs_save;
                //     for (uint8 r : needs_save)
                //         last_needs_save.push(r);

                //     allocate_register(func_asm);

                //     if (last_size != spilled_stack_size)
                //     {
                //         printf("last_size != spilled_stack_size\n");
                //         exit(91);
                //     }
                //     // if (last_needs_save.len != needs_save.len) {
                //     //     printf("last_needs_save.len != needs_save.len\n");
                //     //     exit(92);
                //     // } else {
                //     //     for(uint8 r : needs_save) {
                //     //         if (last_needs_save.find(r) == -1) {
                //     //             printf("cannot find in last_needs_save\n");
                //     //             exit(93);
                //     //         }
                //     //     }
                //     // }
                // }
            }
            if (done_post_work || need_to_legalize_imm) {
                // done, replace all vreg with actual regs
                for (auto b : func_asm->mbs)
                {
                    for (auto I = func_asm->mbs[b->i]->inst; I; I = I->next)
                    {
                        switch (I->tag)
                        {

                        case MI_CLZ:
                        {
                            color_register(((MI_Clz *)I)->dst);
                            color_register(((MI_Clz *)I)->operand);
                        }
                        break;

                        case MI_MOVE:
                        {
                            color_register(((MI_Move *)I)->dst);
                            color_register(((MI_Move *)I)->src);
                        }
                        break;

                        case MI_BINARY:
                        {
                            color_register(((MI_Binary *)I)->dst);
                            color_register(((MI_Binary *)I)->lhs);
                            color_register(((MI_Binary *)I)->rhs);
                        }
                        break;

                        case MI_COMPARE:
                        {
                            color_register(((MI_Compare *)I)->lhs);
                            color_register(((MI_Compare *)I)->rhs);
                        }
                        break;

                        case MI_SLT:
                        {
                            color_register(((MI_Slt *)I)->dst);
                            color_register(((MI_Slt *)I)->lhs);
                            color_register(((MI_Slt *)I)->rhs);
                        }
                        break;

                        case MI_LOAD:
                        case MI_STORE:
                        {
                            color_register(((MI_Load *)I)->reg);
                            color_register(((MI_Load *)I)->base);
                            color_register(((MI_Load *)I)->offset);
                        }
                        break;

                        case MI_BRANCH:
                        {
                            if (((MI_Branch *)I)->cond != NO_CONDITION)
                            {
                                color_register(((MI_Branch *)I)->lhs);
                                color_register(((MI_Branch *)I)->rhs);
                            }
                        }
                        break;

                        // ----------floating point operation----------
                        case MI_VLOAD:
                        case MI_VSTORE:
                        {
                            auto vmv = ((MI_VLoad *)I);
                            if (vmv->base.tag == REG || vmv->base.tag == VREG)
                            {
                                color_register(vmv->base);
                            }
                            if (vmv->offset.tag == REG || vmv->offset.tag == VREG)
                            {
                                color_register(vmv->offset);
                            }
                        }
                        break;

                        case MI_VCMP_SET:
                        {
                            color_register(((MI_VCmp_Set *)I)->dst);
                        }
                        break;
                        case MI_VCVT:
                        {
                            auto fcvt = (MI_VCvt *)I;
                            if (fcvt->from_type == S32)
                            {
                                color_register(fcvt->src);
                            }
                            if (fcvt->to_type == S32)
                            {
                                color_register(fcvt->dst);
                            }
                        }
                        break;

                        case MI_VMOVE:
                        {
                            if (((MI_VMove *)I)->from_s32)
                            {
                                color_register(((MI_VMove *)I)->src);
                            }
                        }
                        break;
                        case MI_VBINARY:
                        case MI_VPUSH:
                        case MI_VPOP:
                        case MI_RETURN:
                        case MI_FUNC_CALL:
                        case MI_PUSH: // we don't push/pop virtual regs, for now
                        case MI_POP:
                            break;

                        default:
                        {
                            cout << "error";
                            exit(65);
                            assert(false);
                        }
                        }
                    }
                }

                printf(">>> Allocated registers:\n");
                for (auto it = color.begin(); it != color.end(); it++)
                {
                    print_operand(it->first);
                    printf(" <- ");
                    print_operand(make_reg(it->second));
                    printf("\n");
                }
                printf("\n");
            }
        }
    }

    void allocate_single_function(Func_Asm *func_asm, bool opt)
    {
        if (opt && func_asm->vreg_count > 3000)
        {

            STACK_RA::stack_ra_on_function(func_asm);
        }
        else
        {
            spilled_stack_size = 0;
            done_post_work = false;
            already_spilled.clear();
            allocate_register(func_asm);
        }
    }

    void s_assign_colors()
    {
        while (!select_stack.empty())
        {
            auto n = select_stack.back();
            select_stack.pop();
            Array<uint8> ok_colors;
            if (in(n, precolored)) continue;

            // assign caller save first
            for (uint8 r = f10; r <= f17; r++)
            { // fa0~fa7
                ok_colors.push(r);
            }
            for (uint8 r = f0; r <= f7; r++)
            { // ft0~ft7
                ok_colors.push(r);
            }
            for (uint8 r = f28; r <= f31; r++)
            { // ft8~ft11
                ok_colors.push(r);
            }

            // assign callee save regs
            for (uint8 r = f8; r <= f9; r++)
            { // fs0~fs1
                ok_colors.push(r);
            }
            for (uint8 r = f18; r <= f27; r++)
            { // fs2~fs11
                ok_colors.push(r);
            }

            // remove adjcent colors(heuristic algorithm)
            for (auto w : adj_list[n])
            {
                auto a = get_alias(w);
                if (in(a, colored_nodes) || in(a, precolored))
                {
                    ok_colors.remove(s_color[a]);
                }
            }
            if (ok_colors.empty())
            {
                spilled_nodes.insert(n);
            }
            else
            {
                colored_nodes.insert(n);
                auto c = ok_colors.front();
                s_color[n] = c;
            }
        }
        for (auto n : coalesced_nodes)
        {
            s_color[n] = s_color[get_alias(n)];
        }
        cout << "my corloring result:" << endl;
        for (auto it : s_color)
        {
            cout << it.first.adr << ": " << it.second << endl;
        }
    }

    void s_allocate_register(Func_Asm *func_asm)
    {
        precolored.clear();
        initial.clear();
        simplify_worklist.clear();
        freeze_worklist.clear();
        spill_worklist.clear();
        spilled_nodes.clear();
        coalesced_nodes.clear();
        colored_nodes.clear();
        coalesced_moves.clear();
        s_coalesced_moves.clear();
        constrained_moves.clear();
        s_constrained_moves.clear();
        frozen_moves.clear();
        s_frozen_moves.clear();
        select_set.clear();
        worklist_moves.clear();
        s_worklist_moves.clear();
        active_moves.clear();
        s_active_moves.clear();
        adj_set.clear();
        adj_list.clear();
        degree.clear();
        move_list.clear();
        s_move_list.clear();
        alias.clear();
        color.clear();
        s_color.clear();
        use_def_count.clear();
        loop_depth.clear();

        select_stack.len = 0;
        needs_save.len = 0;

        s_analyse_liveness(func_asm);
        // @Check: precolored register?
        for (int i = 0; i < SREG_COUNT; i++)
        {
            auto r = make_sreg(i);
            precolored.insert(r);
            s_color[r] = i;
        }

        for (int i = 0; i < func_asm->vsreg_count; i++)
        {
            initial.insert(make_vsreg(i));
        }

        printf(">>> Building interference graph for coloring...\n");
        s_build(func_asm);
        printf(">>> done building\n");
        s_make_worklist();
        printf(">>> worklist done\n");
        do
        {
            if (!simplify_worklist.empty())
                s_simplify();
            else if (!s_worklist_moves.empty())
                s_coalesce();
            else if (!freeze_worklist.empty())
                s_freeze();
            else if (!spill_worklist.empty())
                s_select_spill();
        } while (
            !simplify_worklist.empty() ||
            !s_worklist_moves.empty() ||
            !freeze_worklist.empty() ||
            !spill_worklist.empty());

        printf("assigning\n");
        s_assign_colors();
        printf("assigned\n");

        printf("\n");
        if (!spilled_nodes.empty())
        {
            // rewrite_program(spilled_nodes);

            // rewrite program
            // insert stores and restores for each spilled node
            for (MOperand n : spilled_nodes)
            {

                already_spilled.insert(n);

                printf("spilling nodes: ");
                print_operand(n);
                printf("\n");
                spilled_stack_size += 8; // Alignment

                for (auto bb : func_asm->mbs)
                {
                    for (auto I = bb->inst; I; I = I->next)
                    {
                        auto defs = s_get_defs(I);
                        auto uses = s_get_uses(I, func_asm->has_return_value);

                        bool inserted_store = false;
                        bool inserted_use = false;
                        // insert store after def
                        if (defs.find(n) != -1)
                        {
                            auto str = new MI_VStore;
                            str->mem_tag = MEM_SAVE_SPILL;
                            str->reg = make_vsreg(func_asm->vsreg_count++);
                            str->base = make_reg(sp);
                            str->offset = make_imm(-(spilled_stack_size));
                            str->is_rv64 = true;
                            already_spilled.insert(str->reg);

                            s_replace_defs(I, n, str->reg);
                            insert((MI *)str, I->next); // @Last?
                            I = I->next;                // jump past newly inserted str
                            inserted_store = true;
                        }

                        // insert restore before use
                        if (uses.find(n) != -1)
                        {
                            auto ldr = new MI_VLoad;
                            ldr->mem_tag = MEM_LOAD_SPILL;
                            ldr->reg = make_vsreg(func_asm->vsreg_count++);
                            ldr->base = make_reg(sp);
                            ldr->offset = make_imm(-(spilled_stack_size));
                            ldr->is_rv64 = true;
                            already_spilled.insert(ldr->reg);
                            s_replace_uses(I, n, ldr->reg);
                            // @TODO: get uses based on arg count
                            insert((MI *)ldr, I);
                            inserted_use = true;
                        }
                        assert(!(inserted_use && inserted_store));
                        if (inserted_use && inserted_store)
                        {
                            exit(64);
                        }
                    }
                    // bb->erase_marked_values();
                }
            }

            printf(">>> spilled stack size: %d\n", spilled_stack_size);
            printf(">>> after spilling: \n");
            print_function_asm(func_asm);
            printf("\n");

            s_allocate_register(func_asm);
        }
        else
        {

            // @HACK
            // we need to know needs_save to setup the stack
            // which is computed by color_register
            // which runs after we setup the stack...
            for (auto bb : func_asm->mbs)
            {
                for (auto I = bb->inst; I; I = I->next)
                {
                    for (auto d : s_get_defs(I))
                    {
                        s_color_register(d);
                    }
                }
            }

            if (!done_post_work)
            {
                done_post_work = true;

                func_asm->stack_size += spilled_stack_size;
                func_asm->sreg_needs_save = needs_save.size();
                func_asm->sreg_spill_size += spilled_stack_size;

                printf(">>> registers that need to save: ");

                for (auto r : needs_save)
                {
                    printf("f%d ", r);
                }
                printf("\n");

                // allocate stack space for argument passing
                int max_arg_count = 0;
                int stack_arg_size = 0;
                for (auto mb : func_asm->mbs)
                {
                    for (auto I = mb->inst; I; I = I->next)
                    {
                        if (I->tag == MI_FUNC_CALL)
                        {
                            auto call = (MI_Func_Call *)I;
                            if (call->arg_count > max_arg_count)
                            {
                                max_arg_count = call->arg_count;
                            }
                            if (call->arg_stack_size > stack_arg_size)
                            {
                                stack_arg_size = call->arg_stack_size;
                            }
                        }
                    }
                }

                // if (max_arg_count > 4)
                // {
                //     func_asm->stack_size += stack_arg_size;
                // }

                // cout << ">>> " << func_asm->name << "  " << func_asm->stack_size << endl;

                // insert add/sub to set up or destroy the stack
                if (func_asm->stack_size != 0)
                {
                    auto alloc_stack = new MI_Binary(BINARY_SUBTRACT, make_reg(sp), make_reg(sp), make_imm(func_asm->stack_size));
                    alloc_stack->is_rv64 = true;

                    auto start_inst = func_asm->mbs[0]->inst;
                    // allocate stack after saving callee-save regs
                    while (start_inst != NULL && start_inst->tag == MI_PUSH) {
                        start_inst = start_inst->next;
                    }
                    insert((MI *)alloc_stack, start_inst);

                    for (auto bb : func_asm->mbs)
                    {
                        if (bb->last_inst && bb->last_inst->tag == MI_RETURN)
                        {
                            auto last_inst = bb->last_inst;
                            auto destroy_stack = new MI_Binary(BINARY_ADD, make_reg(sp), make_reg(sp), make_imm(func_asm->stack_size));
                            destroy_stack->is_rv64 = true;

                            // destroy stack before restoring callee-save regs
                            printf("last_instr->prev type: %d\n", last_inst->prev->tag);
                            while (last_inst != NULL && last_inst->prev != NULL 
                                && last_inst->prev->tag == MI_POP) {
                                last_inst = last_inst->prev;
                            }
                            insert((MI *)destroy_stack, last_inst);
                        }
                    }
                }

                // convert fp to sp
                // now func_asm->stack_size = size of (local arrays + spilled vars + max calling arguments);
                // not including the callee-save registers' space

                // fixup sp when calculating base address for local arrays
                for (auto base : func_asm->local_array_bases)
                {
                    assert(base->tag == MI_BINARY);
                    auto sub = (MI_Binary *)base;
                    assert(sub->op == BINARY_SUBTRACT);
                    assert(sub->lhs == make_reg(sp));
                    assert(sub->rhs.tag == IMM);

                    int32 offset_relative_to_sp = func_asm->stack_size - sub->rhs.value;
                    // @TODO replace with a mov directly
                    // if offset relative to sp is 0

                    sub->op = BINARY_ADD;
                    sub->lhs = make_reg(sp);
                    sub->rhs = make_imm(offset_relative_to_sp);
                    sub->lhs.size = 8;
                    sub->is_rv64 = true;
                }

                int save_stack_size = (func_asm->reg_needs_save + func_asm->sreg_needs_save) * 8;
                int all_spill_stack_size = func_asm->reg_spill_size + func_asm->sreg_spill_size;
                for (auto bb : func_asm->mbs)
                {
                    for (auto I = bb->inst; I; I = I->next)
                    {

                        
                        // int save_stack_size = 0;
                        if (I->tag == MI_LOAD)
                        {
                            auto ldr = (MI_Load *)I;
                            if (ldr->mem_tag == MEM_LOAD_SPILL)
                            {
                                int32 offset_value = ((ldr->offset.tag == ERRORTYPE) ? 0 : ldr->offset.value);
                                int32 offset_relative_to_sp = all_spill_stack_size + stack_arg_size + offset_value;
                                ldr->base.value = sp;
                                ldr->offset = make_imm(offset_relative_to_sp);
                            }
                            else if (ldr->mem_tag == MEM_LOAD_ARG)
                            {
                                int32 offset_value = ((ldr->offset.tag == ERRORTYPE) ? 0 : ldr->offset.value);
                                int32 offset_relative_to_sp = func_asm->stack_size + offset_value + save_stack_size;
                                ldr->base.value = sp;
                                ldr->offset = make_imm(offset_relative_to_sp);
                            }
                        }
                        else if (I->tag == MI_STORE)
                        {
                            auto str = (MI_Store *)I;
                            if (str->mem_tag == MEM_SAVE_SPILL)
                            {
                                int32 offset_value = ((str->offset.tag == ERRORTYPE) ? 0 : str->offset.value);
                                int32 offset_relative_to_sp = all_spill_stack_size + stack_arg_size + offset_value;
                                str->base.value = sp;
                                str->offset = make_imm(offset_relative_to_sp);
                            }
                        }
                        else if (I->tag == MI_VLOAD)
                        {
                            auto ldr = (MI_VLoad *)I;
                            if (ldr->mem_tag == MEM_LOAD_SPILL)
                            {
                                int32 offset_value = ((ldr->offset.tag == ERRORTYPE) ? 0 : ldr->offset.value);
                                int32 offset_relative_to_sp = all_spill_stack_size - func_asm->reg_spill_size + stack_arg_size + offset_value;
                                ldr->base.value = sp;
                                ldr->offset = make_imm(offset_relative_to_sp);
                            }
                            else if (ldr->mem_tag == MEM_LOAD_ARG)
                            {
                                int32 offset_value = ((ldr->offset.tag == ERRORTYPE) ? 0 : ldr->offset.value);
                                int32 offset_relative_to_sp = func_asm->stack_size + offset_value + save_stack_size;
                                ldr->base.value = sp;
                                ldr->offset = make_imm(offset_relative_to_sp);
                            }
                        }
                        else if (I->tag == MI_VSTORE)
                        {
                            auto str = (MI_VStore *)I;
                            if (str->mem_tag == MEM_SAVE_SPILL)
                            {
                                int32 offset_value = ((str->offset.tag == ERRORTYPE) ? 0 : str->offset.value);
                                int32 offset_relative_to_sp = all_spill_stack_size - func_asm->reg_spill_size + stack_arg_size + offset_value;
                                str->base.value = sp;
                                str->offset = make_imm(offset_relative_to_sp);
                            }
                        }
                    }
                }

                // insert push/pops to save these registers
                if (!needs_save.empty())
                {
                    // insert push in the start block
                    std::sort(needs_save.begin(), needs_save.end());
                    auto store = new MI_VPush;
                    for (auto r : needs_save)
                    {
                        store->operands.push(make_sreg(r));
                    }
                    insert((MI *)store, func_asm->mbs[0]->inst);

                    // restore registers at exit

                    for (auto bb : func_asm->mbs)
                    {
                        if (bb->last_inst && bb->last_inst->tag == MI_RETURN)
                        {
                            auto restore = new MI_VPop;
                            for (auto r : needs_save)
                            {
                                restore->operands.push(make_sreg(r));
                            }
                            insert((MI *)restore, bb->last_inst);
                        }
                    }
                }

                // check imm
                bool need_to_legalize_imm = false;
                for (auto mb : func_asm->mbs)
                {
                    for (auto I = mb->inst; I; I = I->next)
                    {
                        auto uses = get_uses(I, func_asm->has_return_value);

                        bool inst_need_legalize = false;
                        MOperand use_of_imm;

                        if (I->tag == MI_LOAD)
                        {
                            auto load = (MI_Load *)I;
                            if (load->base.tag == IMM)
                            {
                                goto done;
                            }
                        }

                        if (I->tag == MI_LOAD || I->tag == MI_STORE)
                        {
                            auto load_or_store = (MI_Load *)I;
                            if (load_or_store->offset.tag == IMM)
                            {
                                use_of_imm = load_or_store->offset;
                                inst_need_legalize = !can_be_imm12(use_of_imm.value);
                                goto done;
                            }
                        }

                        if (I->tag == MI_BINARY)
                        {
                            auto bi = (MI_Binary *)I;
                            if (bi->rhs.tag == IMM)
                            {
                                use_of_imm = bi->rhs;
                                inst_need_legalize = !can_be_imm12(use_of_imm.value);
                                goto done;
                            }
                        }

                    done:;

                        need_to_legalize_imm |= inst_need_legalize;

                        if (inst_need_legalize)
                        {
                            printf("%d cannot be imm!\n", use_of_imm.value);

                            // allocate a temp vreg to load the constant from literal pool
                            auto temp = make_vreg(func_asm->vreg_count++);
                            MI_Load *ldr = emit_load_of_constant(temp, use_of_imm.value);
                            insert(ldr, I);

                            // RISC-V do not allow 'sw a1,a0(s2)'('a0'=temp)
                            // we should split it into 'add' and 'sw'
                            if (I->tag == MI_LOAD || I->tag == MI_STORE)
                            {
                                auto load_or_store = (MI_Load *)I;
                                auto new_base = make_vreg(func_asm->vreg_count++);
                                MI_Binary *bi_add = new MI_Binary(BINARY_ADD, new_base, load_or_store->base, temp);
                                bi_add->is_rv64 = true;
                                insert(bi_add, I);
                                replace_uses(load_or_store, use_of_imm, make_imm(0));
                                replace_uses(load_or_store, load_or_store->base, new_base);
                            }
                            else
                            {
                                replace_uses(I, use_of_imm, temp);
                            }
                        }
                    }
                }

                if (need_to_legalize_imm)
                {
                    printf(">>> need to legalize imm!\n");
                    // print_function_asm(func_asm);

                    // int32 last_size = spilled_stack_size;
                    // Array<uint8> last_needs_save;
                    // for (uint8 r : needs_save)
                    //     last_needs_save.push(r);

                    allocate_register(func_asm);

                    // if (last_size != spilled_stack_size)
                    // {
                    //     printf("last_size != spilled_stack_size\n");
                    //     exit(91);
                    // }
                    // if (last_needs_save.len != needs_save.len) {
                    //     printf("last_needs_save.len != needs_save.len\n");
                    //     exit(92);
                    // } else {
                    //     for(uint8 r : needs_save) {
                    //         if (last_needs_save.find(r) == -1) {
                    //             printf("cannot find in last_needs_save\n");
                    //             exit(93);
                    //         }
                    //     }
                    // }
                }

                // done, replace all vreg with actual regs
                for (auto b : func_asm->mbs)
                {
                    for (auto I = func_asm->mbs[b->i]->inst; I; I = I->next)
                    {
                        switch (I->tag)
                        {

                        case MI_CLZ:
                        {
                            s_color_register(((MI_Clz *)I)->dst);
                            s_color_register(((MI_Clz *)I)->operand);
                        }
                        break;

                        case MI_VMOVE:
                        {
                            s_color_register(((MI_VMove *)I)->dst);
                            if (!((MI_VMove *)I)->from_s32)
                            {
                                s_color_register(((MI_VMove *)I)->src);
                            }
                        }
                        break;

                        case MI_VBINARY:
                        {
                            s_color_register(((MI_VBinary *)I)->dst);
                            s_color_register(((MI_VBinary *)I)->lhs);
                            s_color_register(((MI_VBinary *)I)->rhs);
                        }
                        break;

                        case MI_VCMP_SET:
                        {
                            s_color_register(((MI_VCmp_Set *)I)->lhs);
                            s_color_register(((MI_VCmp_Set *)I)->rhs);
                        }
                        break;

                        case MI_VLOAD:
                        case MI_VSTORE:
                        {
                            s_color_register(((MI_VLoad *)I)->reg);
                            // s_color_register(((MI_Load *) I)->base);
                            // s_color_register(((MI_Load *) I)->offset);
                        }
                        break;
                        case MI_VCVT:
                        {
                            auto fcvt = (MI_VCvt *)I;
                            if (fcvt->from_type == F32)
                            {
                                s_color_register(fcvt->src);
                            }
                            if (fcvt->to_type == F32)
                            {
                                s_color_register(fcvt->dst);
                            }
                        }

                        case MI_VPUSH:
                        case MI_VPOP:

                        case MI_RETURN:
                        case MI_FUNC_CALL:
                        case MI_PUSH: // we don't push/pop virtual regs, for now
                        case MI_POP:
                        case MI_BRANCH:
                        case MI_MOVE:
                        case MI_BINARY:
                        case MI_COMPARE:
                        case MI_SLT:
                        case MI_LOAD:
                        case MI_STORE:
                            break;

                        default:
                        {
                            printf("unknown MI tag: %d\n", I->tag);
                            assert(false);
                        }
                        }
                    }
                }

                printf(">>> Allocated registers:\n");
                for (auto it = s_color.begin(); it != s_color.end(); it++)
                {
                    print_operand(it->first);
                    printf(" <- ");
                    print_operand(make_sreg(it->second));
                    printf("\n");
                }
                printf("\n");
            }
        }
    }

    void s_allocate_single_function(Func_Asm *func_asm, bool opt)
    {
        if (opt && func_asm->vsreg_count > 3000)
        {

            STACK_RA::stack_ra_on_function(func_asm);
        }
        else
        {
            spilled_stack_size = 0;
            done_post_work = false;
            already_spilled.clear();
            cout << "ready!!" << endl;
            s_allocate_register(func_asm);
        }
    }
}

void register_allocation(Program_Asm *program_asm, bool opt)
{
    printf(">>> Allocating registers...\n\n");
    for (auto f : program_asm->functions)
    {
        printf(">>> Allocating registers for function %s\n", f->name.c_str());
        RA::allocate_single_function(f, opt);
    }
    printf("------------------------------------------");
}

void sregister_allocation(Program_Asm *program_asm, bool opt)
{
    printf(">>> Allocating sregisters...\n\n");
    for (auto f : program_asm->functions)
    {
        printf(">>> Allocating sregisters for function %s\n", f->name.c_str());
        RA::s_allocate_single_function(f, opt);
    }
}